@@ -1,6 +1,6 @@
 /*
 **  Str - String Library
-**  Copyright (c) 1999-2000 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 1999-2001 Ralf S. Engelschall <rse@engelschall.com>
 **
 **  This file is part of Str, a string handling and manipulation 
 **  library which can be found at http://www.engelschall.com/sw/str/.
@@ -83,7 +83,7 @@
  * FreeBSD version which took it from GNU libc-4.6.27 and modified it
  * to be thread safe. The whole code was finally cleaned up, stripped
  * and extended by Ralf S. Engelschall for use inside the Str library.
- * Especially and Apache and network specific kludges were removed again
+ * Especially any Apache and network specific kludges were removed again
  * and instead the formatting engine now can be extended by the caller
  * on-the-fly.
  */
@@ -112,10 +112,19 @@ typedef unsigned long      u_quad_int;
 #define NUM_BUF_SIZE    512
 #define NDIG            80
 
+/* compatibility */
+#if !defined(HAVE_ISNAN)
+#define isnan(d) (0)
+#endif
+#if !defined(HAVE_ISINF)
+#define isinf(d) (0)
+#endif
+
 /* 
- * convert string to decimal. the number of digits is specified by
- * ndigit decpt is set to the position of the decimal point sign is set
- * to 0 for positive, 1 for negative. buf must have at least NDIG bytes.
+ * Convert decimal number to its string representation. The number of
+ * digits is specified by ndigit decpt is set to the position of the
+ * decimal point sign is set to 0 for positive, 1 for negative. buf must
+ * have at least NDIG bytes.
  */
 
 #define str_ecvt(arg,ndigits,decpt,sign,buf) \
@@ -152,9 +161,9 @@ str_cvt(
     /* Do integer part */
     if (fi != 0) {
         p1 = &buf[NDIG];
-        while (fi != 0) {
+        while (fi != 0 && p1 > &buf[0]) {
             fj = modf(fi / 10, &fi);
-            *--p1 = (int) ((fj + .03) * 10) + '0';
+            *--p1 = (int)((fj + .03) * 10) + '0';
             r2++;
         }
         while (p1 < &buf[NDIG])
@@ -200,7 +209,7 @@ str_cvt(
         }
     }
     *p = NUL;
-    return (buf);
+    return buf;
 }
 
 static char *
@@ -265,7 +274,7 @@ str_gcvt(
     if (p2[-1] == '.' && !altform)
         p2--;
     *p2 = NUL;
-    return (buf);
+    return buf;
 }
 
 /*
@@ -293,10 +302,10 @@ str_gcvt(
  */
 #define NUM(c) ((c) - '0')
 #define STR_TO_DEC(str, num) {    \
-    num = NUM( *str++ );          \
+    num = NUM(*str++);            \
     while (str_isdigit(*(str))) { \
         num *= 10 ;               \
-        num += NUM( *str++ ) ;    \
+        num += NUM(*str++) ;      \
     }                             \
 }
      
@@ -328,8 +337,8 @@ str_gcvt(
  * Increase length. Set the has_prefix flag.
  */
 #define PREFIX(str, length, ch) \
-    *--str = ch; \
-    length++;    \
+    *--str = ch;                \
+    length++;                   \
     has_prefix = TRUE
 
 /*
@@ -863,41 +872,61 @@ str_vformat(
                 case 'e':
                 case 'E':
                     fp_num = va_arg(ap, double);
-                    /* We use &num_buf[ 1 ], so that we have room for the sign */
-                    s = conv_fp(*fmt, fp_num, alternate_form,
-                                (adjust_precision == FALSE) ?  FLOAT_DIGITS : precision, 
-                                &is_negative, &num_buf[1], &s_len);
-                    if (is_negative)
-                        prefix_char = '-';
-                    else if (print_sign)
-                        prefix_char = '+';
-                    else if (print_blank)
-                        prefix_char = ' ';
+                    if (isnan(fp_num)) {
+                        s = "NaN";
+                        s_len = 3;
+                    }
+                    else if (isinf(fp_num)) {
+                        s = "Inf";
+                        s_len = 3;
+                    }
+                    else {
+                        /* use &num_buf[1], so that we have room for the sign */
+                        s = conv_fp(*fmt, fp_num, alternate_form,
+                                    (adjust_precision == FALSE) ? FLOAT_DIGITS : precision, 
+                                    &is_negative, &num_buf[1], &s_len);
+                        if (is_negative)
+                            prefix_char = '-';
+                        else if (print_sign)
+                            prefix_char = '+';
+                        else if (print_blank)
+                            prefix_char = ' ';
+                    }
                     break;
 
                 /* Double Floating Point (style 2) */
                 case 'g':
                 case 'G':
-                    if (adjust_precision == FALSE)
-                        precision = FLOAT_DIGITS;
-                    else if (precision == 0)
-                        precision = 1;
-                    /* We use &num_buf[ 1 ], so that we have room for the sign */
-                    s = str_gcvt(va_arg(ap, double), precision, &num_buf[1],
-                                 alternate_form);
-                    if (*s == '-')
-                        prefix_char = *s++;
-                    else if (print_sign)
-                        prefix_char = '+';
-                    else if (print_blank)
-                        prefix_char = ' ';
-                    s_len = str_len(s);
-                    if (alternate_form && (q = str_span(s, 0, ".", 0)) == NULL) {
-                        s[s_len++] = '.';
-                        s[s_len] = NUL; /* delimit for following str_span() */
+                    fp_num = va_arg(ap, double);
+                    if (isnan(fp_num)) {
+                        s = "NaN";
+                        s_len = 3;
+                    }
+                    else if (isinf(fp_num)) {
+                        s = "Inf";
+                        s_len = 3;
+                    }
+                    else {
+                        if (adjust_precision == FALSE)
+                            precision = FLOAT_DIGITS;
+                        else if (precision == 0)
+                            precision = 1;
+                        /* use &num_buf[1], so that we have room for the sign */
+                        s = str_gcvt(fp_num, precision, &num_buf[1], alternate_form);
+                        if (*s == '-')
+                            prefix_char = *s++;
+                        else if (print_sign)
+                            prefix_char = '+';
+                        else if (print_blank)
+                            prefix_char = ' ';
+                        s_len = str_len(s);
+                        if (alternate_form && (q = str_span(s, 0, ".", 0)) == NULL) {
+                            s[s_len++] = '.';
+                            s[s_len] = NUL; /* delimit for following str_span() */
+                        }
+                        if (*fmt == 'G' && (q = str_span(s, 0, "e", 0)) != NULL)
+                            *q = 'E';
                     }
-                    if (*fmt == 'G' && (q = str_span(s, 0, "e", 0)) != NULL)
-                        *q = 'E';
                     break;
 
                 /* Single Character */
@@ -1077,3 +1106,26 @@ int str_format(char *s, str_size_t n, const char *fmt, ...)
     return rv;
 }
 
+#if 0 /* still don't know whether it fits into API */
+char *str_aprintf(const char *fmt, ...)
+{
+    va_list ap;
+    char *s;
+    int rv;
+
+    va_start(ap, fmt);
+    rv = str_format_va(NULL, -1, fmt, ap);
+    va_end(ap);
+    if (rv == -1)
+        return NULL;
+    if ((s = malloc(rv+1)) == NULL)
+        return NULL;
+    va_start(ap, fmt);
+    rv = str_format_va(s, rv+1, fmt, ap);
+    va_end(ap);
+    if (rv == -1)
+        return NULL;
+    return s;
+}
+#endif 
+
