@@ -120,6 +120,8 @@ hash_destroy(void)
         while (he != NULL) {
             ohe = he;
             he = he->next;
+            if (ohe->key != NULL)
+                free(ohe->key);
             free(ohe);
         }
     }
@@ -268,8 +270,8 @@ int str_parse(const char *string, const char *pattern, ...)
 }
 int str_parse_va(const char *string, const char *pattern, va_list ap)
 {
-    pcre *p_pcre;
-    pcre_extra *p_pcre_extra;
+    pcre *p_pcre = NULL;
+    pcre_extra *p_pcre_extra = NULL;
     const char *match_ptr; 
     int match_len;
     int match_opt;
@@ -415,10 +417,10 @@ int str_parse_va(const char *string, const char *pattern, va_list ap)
     if (cap_num > 0) {
         cap_len = (cap_num+1)*3;
         if ((cap_vec = (int *)malloc(cap_len*sizeof(int))) == NULL) {
-            if (!match_once) {
+            if (p_pcre != NULL)
                 free(p_pcre);
+            if (p_pcre_extra != NULL)
                 free(p_pcre_extra);
-            }
             return -1;
         }
     }
@@ -430,10 +432,10 @@ int str_parse_va(const char *string, const char *pattern, va_list ap)
     if (n < 0) {
         if (cap_vec != NULL)
             free(cap_vec);
-        if (!match_once) {
+        if (p_pcre != NULL)
             free(p_pcre);
+        if (p_pcre_extra != NULL)
             free(p_pcre_extra);
-        }
         if (n == PCRE_ERROR_NOMATCH)
             return 0;
         return -1;
@@ -564,8 +566,15 @@ int str_parse_va(const char *string, const char *pattern, va_list ap)
         l = str_vformat(&sf, buf_ptr, ap);
 
         /* allocate output buffer */
-        if ((*cpp = (char *)malloc(l+1)) == NULL)
+        if ((*cpp = (char *)malloc(l+1)) == NULL) {
+            if (cap_vec != NULL)
+                free(cap_vec);
+            if (p_pcre != NULL)
+                free(p_pcre);
+            if (p_pcre_extra != NULL)
+                free(p_pcre_extra);
             return -1; /* XXX */
+        }
 
         /* finally expand the substitutions string into output buffer */
         sf.curpos    = *cpp;
@@ -586,10 +595,10 @@ int str_parse_va(const char *string, const char *pattern, va_list ap)
     /* cleanup */
     if (cap_vec != NULL)
         free(cap_vec);
-    if (!match_once) {
+    if (p_pcre != NULL)
         free(p_pcre);
+    if (p_pcre_extra != NULL)
         free(p_pcre_extra);
-    }
     /* return success */
     return 1;
 }
